// Copyright (c) 2014 Tel-Aviv University (Israel).
// All rights reserved.
//
// This file is part of ACG (www.acg.cs.tau.ac.il).
// You can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// Author(s): Michal Kleinbort <balasmic@post.tau.ac.il>
//            Oren Salzman <orenzalz@post.tau.ac.il >          

#ifndef RTG_UTILS_H__RTG
#define RTG_UTILS_H__RTG

#include <boost/math/constants/constants.hpp>
#include <boost/foreach.hpp>
#include <vector>

namespace RTG
{
    /// \brief Class of static methods
    class RTG_utils
    {
    public:

        /** \brief returns the radius for r-NN
            \params based on the number
                    n of points, the ratio between the free space volume and c-space volume,
                    and the dimension d
                    mu is the free space volume out of the c-space which is 1
	    
            @par Based on:
            L. Janson and M. Pavone,
            Fast Marching Trees: a Fast Marching Sampling-Based Method for Optimal
            Motion Planning in Many Dimensions - Extended Version,
            <em>CoRR</em>, abs/1306.3532, 2013.
        */
        static double calculate_radius(int n, double d, double mu = 1.0, double vol = 1.0)
	    {
		    double z_d = get_sphere_volume(d);
		    double delta = 1/d;
		    double v = mu * vol / z_d;
		    double gamma = 2.2 * std::pow(delta, delta) * std::pow(v, delta);
            double m = std::log ((double)n) / n;
		    return gamma * std::pow (m, delta);
	    }

        /// \brief computes the volume of the bounding box of a given point-set
        template<typename _T>
        static double get_bbox_volume(std::vector<_T>& points, unsigned int dim)
        {
            if (points.size() == 0)
                return 0.0;

            //find the min and max coord vals for each axis
            _T min_coords = points[0]; //stores the min coord for each axis
            _T max_coords = points[0]; //stores the max coord for each axis
       
            BOOST_FOREACH (_T& p, points)
            {
                for (std::size_t i(0); i < dim; ++i)
                {
                    if (p[i] < min_coords[i])
                        min_coords[i] = p[i];
                    if (p[i] > max_coords[i])
                        max_coords[i] = p[i];
                }
            }

            //compute the volume of the bbox
            double vol = 1.0;
            for (std::size_t i(0); i < dim; ++i)
            {
                vol *= (max_coords[i] - min_coords[i]);
            }

            return vol;
        }
    
    private:
	    /// \brief computes the volume of the d-dimensional unit sphere
        static double get_sphere_volume(double d)
	    {
		    if (d == 0)
			    return 1;
		    if (d == 1)
			    return 2;
		    const double pi = boost::math::constants::pi<double>();
		    return ( (2 * pi) / d ) * get_sphere_volume(d-2);
	    }

    }; //RTG_utils

} //namespace RTG

#endif //RTG_UTILS_H__RTG
