// Copyright (c) 2014 Tel-Aviv University (Israel).
// All rights reserved.
//
// This file is part of ACG (www.acg.cs.tau.ac.il).
// You can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation,
// either version 3 of the License, or (at your option) any later version.
//
// Licensees holding a valid commercial license may use this file in
// accordance with the commercial license agreement provided with the software.
//
// This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
// WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
//
// Author(s): Michal Kleinbort <balasmic@post.tau.ac.il>
//            Oren Salzman <orenzalz@post.tau.ac.il >          

#ifndef NEAREST_NEIGHBORS_RTG_H__RTG
#define NEAREST_NEIGHBORS_RTG_H__RTG

#include <vector>
#include <unordered_set>
#include <map>

#include <boost/unordered_map.hpp>

#include "Random_utils.h" 
#include "Random_matrix_gen.h"
#include "pairs.h"
#include "vec_of_intsets_pairs_struct.h"
#include "Point_d.h"
#include "distance_functors.h"
#include "single_grid.h"

namespace RTG
{
    /** \brief Randomly Translated Grids, a data
        structure for nearest neighbor search.
		Especially for all r-nearest neighbors.
        
        @par Based on:
           D. Aiger, H. Kaplan, and M. Sharir,
           Reporting Neighbors in High-Dimensional Euclidean Space,
           <em>SIAM J. Comput.</em>, vol. 43(4), pp. 1363-1395, 2014.
           DOI: <a href="http://epubs.siam.org/doi/abs/10.1137/12089867X">10.1137/12089867X</a><br>
    */
    template<typename _PointDT = Point_d_,  
			 typename _SqrdDistanceFuncT = EuclideanSqrdDistanceFunc<_PointDT> >
    class NearestNeighborsRTG 
    {
     public:
        typedef _PointDT                                PointD;
        typedef _SqrdDistanceFuncT	                    SqrdDistanceFunc;
		typedef SingleGrid<PointD, SqrdDistanceFunc >   RTGSingleGrid; 

        /// \brief Constructor
        NearestNeighborsRTG(double r, double cell_size, unsigned int grid_num, 
                            unsigned int num_samples, unsigned int dim, 
                            SqrdDistanceFunc* pDistFunc = NULL,
                            PairsStruct* pairs = NULL,
                            bool all_nn_only = true, int seed = 1)
         : r_sq_(r*r), c_(cell_size), m_(grid_num),
           dim_(dim), seed_(seed), all_nn_only_(all_nn_only), rand_(NULL), 
           done_pairs_search_(false), with_rotation_(false), pSqrDistFunc_(pDistFunc),
           bDeleteDistFunc_(false), pairs_(pairs), bDeletePairsStruct_(false)
        {
			rand_ = new Random_utils<PointD>(seed, 0.0, cell_size);
            
            if (pSqrDistFunc_ == NULL)
            {
                BOOST_ASSERT_MSG((boost::is_same<SqrdDistanceFunc, EuclideanSqrdDistanceFunc<PointD> >::value), 
                                 "\n\nERROR MSG: If sqrdDistFunc is not given, the SqrdDistanceFunc type must be EuclideanSqrdDistanceFunc");
    
                pSqrDistFunc_ = new SqrdDistanceFunc(dim_);
                bDeleteDistFunc_ = true;
            }
            if (pairs_ == NULL)
            {
                pairs_ = new VecOfIntSetsPairsStruct(num_samples, all_nn_only);
                bDeletePairsStruct_ = true;
            }
        }

        /// \brief Distructor
        ~NearestNeighborsRTG(void)
        {
           if (rand_ != NULL)
			   delete rand_;

		   rand_ = NULL;
           pairs_->clear();
		   points_.clear();
           if (bDeleteDistFunc_)
                delete pSqrDistFunc_;
           if (bDeletePairsStruct_)
                delete pairs_;
           return;
        }   
        
        void set_pairs_struct(PairsStruct* pairs)
        {
            if (pairs_ != NULL)
            {
                pairs_->clear();
                pairs_ = NULL;
            }
            pairs_ = pairs;
            return;
        }

        /// \brief Adds a data point into the existing array of points
        void add(const PointD &data)
        {
            points_.push_back(data);
            return;
        }
    
        /// \brief Adds a vector of points into the existing array of points
        void add(const std::vector<PointD> &data)
        {
			points_.reserve(points_.size() + data.size());
			points_.insert(points_.end(), data.begin(), data.end());
			return;
        }

        /// \brief A method retrieving the r-nearest-neighbors of the point 
        ///        whose index in the data points array is data_index.
        ///        The neighbors are inserted into the nbh vector of data points.
        ///        This method is not supported when all_nn_only is set to true.
        void nearestR(int data_index, std::vector<PointD> &nbh)
        {
            BOOST_ASSERT_MSG((!all_nn_only_), 
                             "\n\nERROR MSG: nearestR operation is unsupported");

			if (!done_pairs_search_)
				find_all_pairs();

            //find all r-NN of point i in the pairs structure (using neighbors_of_i)
			std::vector<int> nvec;
			pairs_->neighbors_of_i(data_index, nvec);

            //copy from the temporary vector to the nbh vector 
			for (int i(0); i < nvec.size(); ++i)
			{
				nbh.push_back(points_[nvec[i]]);
			}
            return;
        }

        /// \brief A method retrieving the indices of r-nearest-neighbors of the point 
        ///        whose index in the data points array is data_index.
        ///        The indices of neighbors are inserted into the nbh_ind vector of integers.
        ///        This method is not supported when all_nn_only is set to true.
        void nearestR(int data_index, std::vector<int> &nbh_ind)
        {
            BOOST_ASSERT_MSG((!all_nn_only_), 
                             "\n\nERROR MSG: nearestR operation is unsupported");

			if (!done_pairs_search_)
				find_all_pairs();

            //find all r-NN of point i in the pairs structure (using neighbors_of_i)
			pairs_->neighbors_of_i(data_index, nbh_ind);
            return;
        }


        /// \brief A method retrieving all pairs r-nearest-neighbors.
        ///        The pairs of data points are inserted into the nbh_pairs 
        ///          vector of pairs of points.
        void all_nearestR(std::vector< std::pair<PointD, PointD> > &nbh_pairs)
        {
			if (!done_pairs_search_)
				find_all_pairs();

            //find all r-NN pair in the pairs structure (using all_pairs)
            std::vector<std::pair<int, int> > nvec;
			pairs_->all_pairs(nvec);

            //copy from the temporary vector to nbh_pairs vector
			for (int i(0); i < nvec.size(); ++i)
			{
				std::pair<int, int> pr(nvec[i]);
				nbh_pairs.push_back(std::make_pair(points_[pr.first], points_[pr.second]));
			}
            return;
        }


        /// \brief A method retrieving all pairs of indices of r-nearest-neighbors points.
        ///        The pairs of indices are inserted into the nbh_ind_pairs
        ///          vector of pairs of points indices.
        void all_nearestR(std::vector< std::pair<int, int> > &nbh_ind_pairs)
        {
			if (!done_pairs_search_)
				find_all_pairs();

            //find all r-NN pair in the pairs structure (using all_pairs)
            pairs_->all_pairs(nbh_ind_pairs);
            return;
        }

        /// \brief A method retrieving the number of data points in the structure
        std::size_t size(void) const
        {
            return points_.size();
        }
		
     private:

        /// \brief A method updating the pairs structure 
        ///         with all r-NN pairs.
        ///         It constructs m grids and collects all pairs of points 
        ///         from each grid.
        void find_all_pairs() 
		{
			for (unsigned int i=0; i < m_; ++i)
			{
				//construct new grid
				std::vector<double> shift;
				get_random_shift(dim_, shift);
				
				if (with_rotation_)
				{
				    boost::shared_ptr<Eigen::MatrixXd> pmat = rmg_.get_random_rotation_matrix(dim_);
				    RTGSingleGrid grid(shift, c_, r_sq_, points_, dim_, pmat, pSqrDistFunc_);
				    //update pairs using grid
				    grid.find_all_pairs_single(pairs_);
				}
				else
				{
					RTGSingleGrid grid(shift, c_, r_sq_, points_, dim_, pSqrDistFunc_);
					//update pairs using grid
					grid.find_all_pairs_single(pairs_);
				}
			}
			done_pairs_search_ = true;
		}

        /// \brief A method for selecting a random shift
        ///         which is represented by a random point in dimension dim.
        void get_random_shift(int dim, std::vector<double>& vec) const
		{
			rand_->get_random_vec(dim, vec); 
		}

        
     private:
        
		/// \brief The data structure containing the point set
		std::vector<PointD>			points_;
		/// \brief The squared radius that will be used in the NN search
		double						r_sq_;
		/// \brief The cell size
		double						c_;
		/// \brief The dimension of the points in the point set
		unsigned int				dim_;
		/// \brief The seed used for random number generating
		int	    					seed_;
		/// \brief The random Matrix generator
		RandomMatrixGen             rmg_;
		/// \brief The number of grids
		unsigned int                m_;
		/// \brief A flag indicating whether only an "all pairs" query 
		///  will be supported
		bool                        all_nn_only_;
		/// \brief A pointer for a data structure storing the already detected pairs
        ///        It should be deleted from the outside (where it is allocated)
		PairsStruct*                pairs_;
		/// \brief Random number generator utility object
		Random_utils<PointD>*       rand_;
		/// \brief Flag indicating whether find_all_pairs operation took place
		bool						done_pairs_search_;
		/// \brief Flag indicating whether to use rotation in the tandom transformation of the grid
		bool						with_rotation_;
        /// \brief A pointer to sqr distance retriever functor
        SqrdDistanceFunc*           pSqrDistFunc_;
        /// \brief Flag indicating whether to delete the sqr distance retriever functor
        bool                        bDeleteDistFunc_;
        /// \brief Flag indicating whether to delete the auxiliary pairs structure
        bool                        bDeletePairsStruct_;
    };


} //namespace RTG
#endif //NEAREST_NEIGHBORS_RTG_H__RTG